"""Connection module"""

from __future__ import annotations

import asyncio
import base64
import json
import logging
import urllib.parse
from typing import Any

import aiohttp

from asusrouter import (
    AsusRouter404,
    AsusRouterAuthorizationError,
    AsusRouterConnectionError,
    AsusRouterConnectionTimeoutError,
    AsusRouterError,
    AsusRouterLoginBlockError,
    AsusRouterLoginError,
    AsusRouterResponseError,
    AsusRouterServerDisconnectedError,
    AsusRouterSSLError,
)
from asusrouter.const import (
    AR_API,
    AR_ERROR_CODE,
    AR_USER_AGENT,
    AUTHORIZATION,
    CREDENTIALS,
    DEFAULT_PORT,
    DEFAULT_SLEEP_CONNECTING,
    DEFAULT_SLEEP_RECONNECT,
    ENDPOINT,
    ENDPOINT_LOGIN,
    ENDPOINT_LOGOUT,
    ERROR_STATUS,
    HOOK,
    HTTP,
    HTTPS,
    LOGOUT,
    SUCCESS,
    TRY_AGAIN,
)
from asusrouter.util import converters, parsers

_LOGGER = logging.getLogger(__name__)


class Connection:
    """Create connection"""

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        port: int | None = None,
        use_ssl: bool = False,
        session: aiohttp.ClientSession | None = None,
    ):
        """Properties for connection"""

        self._host: str | None = host
        self._port: int | None = port
        self._username: str | None = username
        self._password: str | None = password
        self._token: str | None = None
        self._headers: dict[str, str] | None = None
        self._session: aiohttp.ClientSession | None = session

        self._device: dict[str, str] = {}
        self._error: bool = False
        self._connecting: bool = False
        self._mute_flag: bool = False
        self._drop: bool = False

        self._http = HTTPS if use_ssl else HTTP
        self._ssl = False

        if self._port is None or self._port == 0:
            self._port = DEFAULT_PORT[self._http]

        self._connected: bool = None

    async def async_run_command(
        self, command: str, endpoint: str = ENDPOINT[HOOK], retry: bool = False
    ) -> dict[str, Any]:
        """Run command. Use the existing connection token, otherwise create new one"""

        if self._drop:
            return {}

        if self._token is None and not retry:
            _LOGGER.debug("No token - connecting and repeating")
            await self.async_connect()
            return await self.async_run_command(command, endpoint, retry=True)
        if self._token is not None:
            try:
                result = await self.async_request(command, endpoint, self._headers)
                return result
            except AsusRouterAuthorizationError as ex:
                if not retry:
                    await self.async_connect()
                    return await self.async_run_command(command, endpoint, retry=True)
                raise ex
            except AsusRouterError as ex:
                raise ex
            except Exception as ex:
                if not retry:
                    await self.async_connect()
                    return await self.async_run_command(command, endpoint, retry=True)
                raise ex
        else:
            _LOGGER.error("Error sending a command to `%s`", endpoint)
            return {}

    async def async_load(self, endpoint: str, retry: bool = False) -> str:
        """Load the page"""

        if self._drop:
            return {}

        if self._token is None and not retry:
            await self.async_connect()
            return await self.async_load(endpoint, retry=True)

        string_body = str()

        try:
            async with self._session.post(
                url=f"{self._http}://{self._host}:{self._port}/{endpoint}",
                headers=self._headers,
                ssl=self._ssl,
            ) as reply:
                string_body = await reply.text()
                if "404 Not Found" in string_body:
                    raise AsusRouter404()
        except (aiohttp.ClientError, asyncio.TimeoutError) as ex:
            if not retry:
                return self.async_load(endpoint, retry=True)
            raise AsusRouterConnectionError(ex) from ex
        except Exception as ex:
            raise ex

        return string_body

    async def async_request(
        self, payload: str, endpoint: str, headers: dict[str, str], retry: bool = False
    ) -> dict[str, Any]:
        """Send a request"""

        if self._drop:
            return {}

        # Wait a bit before just retrying
        if retry:
            await asyncio.sleep(DEFAULT_SLEEP_RECONNECT)

        # Connection process
        if endpoint == ENDPOINT_LOGIN:
            self._connecting = True
        # Sleep if we got here during the connection process
        else:
            while self._connecting:
                await asyncio.sleep(DEFAULT_SLEEP_CONNECTING)

        json_body = {}

        try:
            async with self._session.post(
                url=f"{self._http}://{self._host}:{self._port}/{endpoint}",
                data=urllib.parse.quote(payload),
                headers=headers,
                ssl=self._ssl,
            ) as reply:
                string_body = await reply.text()
                if string_body == str():
                    return {}
                if "404 Not Found" in string_body:
                    raise AsusRouter404()
                json_body = await reply.json()

                # Handle reported errors
                if ERROR_STATUS in json_body:
                    # If got here, we are not connected properly
                    self._connected = False

                    # ERROR CODES
                    error_code = int(json_body[ERROR_STATUS])
                    # Not authorized
                    if error_code == AR_ERROR_CODE[AUTHORIZATION]:
                        raise AsusRouterAuthorizationError("Session is not authorized")
                    # Wrong crerdentials
                    if error_code == AR_ERROR_CODE[CREDENTIALS]:
                        raise AsusRouterLoginError("Wrong credentials")
                    # Too many attempts
                    if error_code == AR_ERROR_CODE[TRY_AGAIN]:
                        raise AsusRouterLoginBlockError(
                            "Too many attempts to login",
                            timeout=converters.int_from_str(
                                json_body["remaining_lock_time"]
                            ),
                        )
                    # Loged out
                    if error_code == AR_ERROR_CODE[LOGOUT]:
                        return {SUCCESS: True}
                    # Unknown error code
                    raise AsusRouterResponseError(
                        f"Unknown error code `{error_code}`, please report it"
                    )

            # If loged in, save the device API data
            if endpoint == ENDPOINT_LOGIN:
                r_headers = reply.headers
                for item in AR_API:
                    if item in r_headers:
                        self._device[item] = r_headers[item]

            # Reset mute_flag on success
            self._mute_flag = False

            return json_body

        # Handle non-JSON replies
        except json.JSONDecodeError:
            if ".xml" in endpoint:
                json_body = parsers.xml(text=string_body)
            else:
                json_body = parsers.pseudo_json(text=string_body, page=endpoint)
            return json_body

        # Raise only if mute_flag not set
        except aiohttp.ClientSSLError as ex:
            if not self._mute_flag:
                raise AsusRouterSSLError() from ex
        except (aiohttp.ServerTimeoutError, asyncio.TimeoutError) as ex:
            if not self._mute_flag:
                raise AsusRouterConnectionTimeoutError() from ex
        # This happens regularly if more connections are established.
        # Repeat before actually raising the exception
        except aiohttp.ServerDisconnectedError as ex:
            if retry:
                raise AsusRouterServerDisconnectedError(ex) from ex
            _LOGGER.debug(
                "Disconnected by the device while quering '%s' with payload '%s'. "
                "Everything should recover by itself. If this warning appears regularly, "
                "you might need to decrease the number of simultaneous connections "
                "to your device",
                endpoint,
                payload,
            )

        # Connection refused or reset by peer -> will repeat
        except (aiohttp.ClientOSError) as ex:
            if endpoint == ENDPOINT_LOGIN and not self._mute_flag:
                raise AsusRouterConnectionError(str(ex)) from ex
            # Mute warning for retries and while connecting
            if not retry and not self._connecting:
                _LOGGER.debug(
                    "Connection refused. Endpoint: %s. Payload: %s.\nException summary: %s",
                    endpoint,
                    payload,
                    str(ex),
                )

        # If it got here, something is wrong. Reconnect and retry
        self._mute_flag = True

        if not retry:
            self._error = True
            _LOGGER.debug("Reconnecting")
            await self.async_cleanup()
            await self.async_connect(retry=True)
        return await self.async_request(
            payload=payload, endpoint=endpoint, headers=headers, retry=True
        )

    async def async_connect(self, retry: bool = False) -> bool:
        """Start new connection to and get new auth token"""

        _LOGGER.debug("Trying to connect")

        _success = False
        self._drop = False

        if self._session is None:
            _LOGGER.debug("No client session provided. Starting a new session")
            self._session = aiohttp.ClientSession()

        auth = f"{self._username}:{self._password}".encode("ascii")
        logintoken = base64.b64encode(auth).decode("ascii")
        payload = f"login_authorization={logintoken}"
        headers = {"user-agent": AR_USER_AGENT}

        _LOGGER.debug("Sending connection request")
        response = await self.async_request(
            payload=payload, endpoint=ENDPOINT_LOGIN, headers=headers, retry=retry
        )
        if "asus_token" in response:
            self._token = response["asus_token"]
            self._headers = {
                "user-agent": AR_USER_AGENT,
                "cookie": f"asus_token={self._token}",
            }
            _LOGGER.debug("Login successful on port `%s`: %s", self._port, self._device)

            self._connected = True
            _success = True
        else:
            _LOGGER.error("Cannot recieve a token from device")

        self._connecting = False

        return _success

    async def async_disconnect(self) -> bool:
        """Close the connection"""

        # Not connected
        if not self._connected:
            _LOGGER.debug("Not connected")
        # Connected
        else:
            try:
                result = await self.async_request("", ENDPOINT_LOGOUT, self._headers)
                if SUCCESS not in result:
                    return False
            except AsusRouterAuthorizationError:
                _LOGGER.debug("Connection was already droped")

        # Clean up
        await self.async_cleanup()

        return True

    async def async_drop_connection(self) -> None:
        """Drop connection"""

        self._drop = True
        await self.async_cleanup()

    async def async_cleanup(self) -> None:
        """Cleanup after logout"""

        _LOGGER.debug("Cleaning up")

        self._connected = False
        self._token = None
        self._headers = None

    async def async_reset_error(self) -> None:
        """Reset error flag"""

        self._error = False
        return

    @property
    def connected(self) -> bool:
        """Connection status"""

        return self._connected

    @property
    def connecting(self) -> bool:
        """Connection progress"""

        return self._connecting

    @property
    def device(self) -> dict[str, str]:
        """Device model and API support levels"""

        return self._device

    @property
    def error(self) -> bool:
        """Report errors"""

        return self._error



"""Compilers module for AsusRouter"""

from __future__ import annotations

from dataclasses import asdict
from typing import Any

from asusrouter import FilterDevice
from asusrouter.const import (
    AR_MAP_RGB,
    ENDHOOKS,
    ENDPOINT,
    ENDPOINT_ARGS,
    ERRNO,
    ERROR_VALUE_TYPE,
    GWLAN,
    HOOK,
    KEY_PARENTAL_CONTROL_MAC,
    KEY_PARENTAL_CONTROL_NAME,
    KEY_PARENTAL_CONTROL_TIMEMAP,
    KEY_PARENTAL_CONTROL_TYPE,
    KEY_PORT_FORWARDING_LIST,
    MAP_NVRAM,
    MAP_OVPN_STATUS,
    MAP_PARENTAL_CONTROL_TYPE,
    NVRAM_GET,
    RANGE_GWLAN,
    RANGE_OVPN_CLIENTS,
    STATE,
    STATUS,
    UNKNOWN,
    VPN,
    VPN_CLIENT,
    WLAN,
    WLAN_TYPE,
)
from asusrouter.dataclass import AsusDevice, ConnectedDevice, PortForwarding
from asusrouter.error import AsusRouterValueError

from .converters import int_from_str


def hook(commands: dict[str, str] | None = None) -> str:
    """Hook compiler"""

    data = str()
    if commands is not None:
        for item, value in commands:
            data += f"{item}({value});"

    return data


def nvram(values: list[str] | str | None = None) -> str:
    """NVRAM compiler"""

    if values is None:
        return str()

    if isinstance(values, str):
        return f"{NVRAM_GET}({values});"

    request = str()
    for value in values:
        request += f"{NVRAM_GET}({value});"

    return request


def rgb(raw: dict[int, dict[str, int]]) -> str:
    """RGB value compiler for LEDG scheme"""

    value = str()

    raw = dict(sorted(raw.items()))

    for led in raw:
        for code in AR_MAP_RGB.values():
            if code in raw[led]:
                value += f"{raw[led][code]},"
            else:
                value += "0,"

    value = value[:-1]

    return value


def vpn_from_devicemap(
    vpn: dict[str, Any], devicemap: dict[str, Any]
) -> dict[str, Any]:
    """Compile devicemap into VPN"""

    if not isinstance(devicemap, dict):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(devicemap, type(devicemap)))

    if not isinstance(vpn, dict):
        vpn = {}

    if VPN in devicemap:
        for num in RANGE_OVPN_CLIENTS:
            key = f"{VPN_CLIENT}{num}"
            if not key in vpn:
                vpn[key] = {}
            if f"{VPN_CLIENT}{num}_{STATE}" in devicemap[VPN]:
                vpn[key][STATUS] = int_from_str(
                    devicemap[VPN][f"{VPN_CLIENT}{num}_{STATE}"]
                )
                if vpn[key][STATUS] == 1 or vpn[key][STATUS] == 2:
                    vpn[key][STATE] = True
                else:
                    vpn[key][STATE] = False
                if vpn[key][STATUS] in MAP_OVPN_STATUS:
                    vpn[key][STATUS] = MAP_OVPN_STATUS[vpn[key][STATUS]]
                else:
                    vpn[key][STATUS] = f"{UNKNOWN} ({vpn[key][STATUS]})"
            if f"{VPN_CLIENT}{num}_{ERRNO}" in devicemap[VPN]:
                vpn[key][ERRNO] = int_from_str(
                    devicemap[VPN][f"{VPN_CLIENT}{num}_{ERRNO}"]
                )

    return vpn


def parental_control(data: dict[str, FilterDevice]) -> dict[str, str]:
    """Compile parental control rules"""

    if not isinstance(data, dict):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(data, type(data)))

    types_lib = {}
    for index, state in MAP_PARENTAL_CONTROL_TYPE.items():
        types_lib[state] = index

    macs = str()
    names = str()
    types = str()
    timemaps = str()

    for rule in data:
        macs += f"{data[rule].mac}>"
        names += f"{data[rule].name}>"
        types += f"{types_lib[data[rule].type]}>"
        timemaps += f"{data[rule].timemap.replace('&#60', '<') if data[rule].timemap is not None else ''}>"

    macs = macs[:-1]
    names = names[:-1]
    timemaps = timemaps[:-1]
    types = types[:-1]

    return {
        KEY_PARENTAL_CONTROL_MAC: macs,
        KEY_PARENTAL_CONTROL_NAME: names,
        KEY_PARENTAL_CONTROL_TIMEMAP: timemaps,
        KEY_PARENTAL_CONTROL_TYPE: types,
    }


def port_forwarding(data: list[PortForwarding]) -> dict[str, str]:
    """Compile port forwarding rules"""

    if not isinstance(data, list):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(data, type(data)))

    result = ""
    for rule in data:
        name = str()
        if rule.name is not None and rule.name != "None":
            name = rule.name
        port_external = str()
        if rule.port_external is not None and rule.port_external != "None":
            port_external = rule.port_external
        ip = str()
        if rule.ip is not None and rule.ip != "None":
            ip = rule.ip
        port = str()
        if rule.port is not None and rule.port != "None":
            port = rule.port
        protocol = str()
        if rule.protocol is not None and rule.protocol != "None":
            protocol = rule.protocol
        ip_external = str()
        if rule.ip_external is not None and rule.ip_external != "None":
            ip_external = rule.ip_external


        result += f"<{name}>{port_external}>{ip}>{port}>{protocol}>{ip_external}>"

    return {
        KEY_PORT_FORWARDING_LIST: result,
    }


def connected_device(
    device: ConnectedDevice, state: dict[str, Any] | None = None
) -> ConnectedDevice:
    """Compile connected device from different sources"""

    if state is None:
        device.online = False
        return device

    values: dict[str, Any] = asdict(device)
    values.update(state)

    return ConnectedDevice(**values)


def endpoint(endpoint: str, device: AsusDevice | dict[str, Any] | None = None) -> str:
    """Compile endpoint string with required arguments"""

    address = ENDPOINT.get(endpoint)
    if not address:
        address = ENDPOINT[HOOK] if endpoint in ENDHOOKS else str()

    if isinstance(device, AsusDevice):
        device = asdict(device)
    if not isinstance(device, dict):
        return address

    address += "?"

    if endpoint in ENDPOINT_ARGS:
        for key in ENDPOINT_ARGS[endpoint]:
            address += f"{ENDPOINT_ARGS[endpoint][key]}={device.get(key, str())}&"

    return address


def update_rec(left: dict[str, Any], right: dict[str, Any] | None = None) -> None:
    """Update dictionary with values of other dictionary"""

    if not right:
        right = {}
    for key, value in right.items():
        if key in left and isinstance(left[key], dict) and isinstance(value, dict):
            update_rec(left[key], value)
        else:
            left[key] = value


def monitor_arg_nvram(wlan: list[str] | None) -> str | None:
    """Compile NVRAM monitor"""

    if not wlan:
        return None

    request = []
    for intf in wlan:
        interface = WLAN_TYPE.get(intf)
        for key in MAP_NVRAM[WLAN]:
            request.append(key.value.format(interface))

        for key in MAP_NVRAM[GWLAN]:
            for gid in RANGE_GWLAN:
                request.append(key.value.format(f"{interface}.{gid}"))

    return nvram(request)



"""AsusRouter constants module"""

from enum import Enum
from typing import Any

from asusrouter.dataclass import Key, SearchKey
from asusrouter.util.converters import (
    bool_from_any,
    bool_or_int,
    datetime_from_str,
    exists_or_not,
    float_from_str,
    int_from_str,
    ovpn_remote_fom_str,
    service_support,
    time_from_delta,
)


### ENUMS
class Merge(str, Enum):
    """Merge enumerator for getting data"""

    ALL = "all"
    ANY = "any"


### API VALUES
AR_USER_AGENT = "asusrouter--DUTUtil-"
AR_API = [
    "Model_Name",
    "AiHOMEAPILevel",
    "Httpd_AiHome_Ver",
]
AR_ERROR_CODE = {
    "authorization": 2,
    "credentials": 3,
    "try_again": 7,
    "logout": 8,
}

### VALUES
ACTION_MODE = "action_mode"
AIMESH = "aimesh"
APPLY = "apply"
APPOBJ = "appobj"
AUTHORIZATION = "authorization"
BLOCK = "block"
BOOTTIME = "boottime"
BRIDGE = "bridge"
CLIENTS = "clients"
CLIENTS_HISTORIC = "clients_historic"
COMMAND = "command"
CONNECTION_TYPE = "connection_type"
CPU = "cpu"
CPU_USAGE = "cpu_usage"
CREDENTIALS = "credentials"
DATA = "data"
DEVICEMAP = "devicemap"
DEVICES = "devices"
DIAG = "diag"
DISABLE = "disable"
DHCP = "dhcp"
ENDPOINTS = "endpoints"
ERROR_STATUS = "error_status"
ERRNO = "errno"
FIRMWARE = "firmware"
FREE = "free"
GUEST = "guest"
GWLAN = "gwlan"
HOOK = "hook"
HTTP = "http"
HTTPS = "https"
INFO = "info"
IP = "ip"
ISO = "iso"
LACP = "lacp"
LAN = "lan"
LED = "led"
LED_VAL = "led_val"
LEDG = "ledg"
LIGHT = "light"
LINK_RATE = "link_rate"
LOAD_AVG = "load_avg"
LOGIN = "login"
LOGOUT = "logout"
MAC = "mac"
MAIN = "main"
MEM = "mem"
MEMORY_USAGE = "memory_usage"
MONITOR = "monitor"
NAME = "name"
NETDEV = "netdev"
NETWORK = "network"
NETWORKMAPD = "networkmapd"
NODE = "node"
NPMCLIENT = "npmclient"
NVRAM = "nvram"
NVRAM_GET = "nvram_get"
ONBOARDING = "onboarding"
ONLINE = "online"
PARENTAL_CONTROL = "parental_control"
PORT = "port"
PORT_FORWARDING = "port_forwarding"
PORTS = "ports"
QTN = "qtn"
RAM = "ram"
RGB = "rgb"
RIP = "rip"
RSSI = "rssi"
RULES = "rules"
RX = "rx"
SERVICE_COMMAND = "rc_service"
SERVICE_MODIFY = "modify"
SERVICE_REPLY = "run_service"
SERVICE_SET_LED = "start_ctrl_led"
SIM = "sim"
STATE = "state"
STATUS = "status"
SUCCESS = "success"
SYS = "sys"
SYSINFO = "sysinfo"
TEMPERATURE = "temperature"
TIME = "time"
TIMEMAP = "timemap"
TIMESTAMP = "timestamp"
TOTAL = "total"
TRY_AGAIN = "try_again"
TX = "tx"
TYPE = "type"
UNKNOWN = "unknown"
UPDATE_CLIENTS = "update_clients"
USAGE = "usage"
USED = "used"
USB = "usb"
VPN = "vpn"
VPN_CLIENT = "vpn_client"
VPN_SERVER = "vpn_server"
WAN = "wan"
WANLINK_STATE = "wanlink_state"
WIRED = "wired"
WLAN = "wlan"
WLAN_2GHZ = "2ghz"
WLAN_5GHZ = "5ghz"
WLAN_5GHZ2 = "5ghz2"
WLAN_6GHZ = "6ghz"

### ASUS DATA TYPES
ETHERNET_PORTS = "ethernet_ports"
PORT_STATUS = "port_status"

### Keys & delimiters
KEY_PARENTAL_CONTROL_MAC = "MULTIFILTER_MAC"
KEY_PARENTAL_CONTROL_NAME = "MULTIFILTER_DEVICENAME"
KEY_PARENTAL_CONTROL_STATE = "MULTIFILTER_ALL"
KEY_PARENTAL_CONTROL_TIMEMAP = "MULTIFILTER_MACFILTER_DAYTIME_V2"
KEY_PARENTAL_CONTROL_TYPE = "MULTIFILTER_ENABLE"
KEY_PORT_FORWARDING_LIST = "vts_rulelist"
KEY_PORT_FORWARDING_STATE = "vts_enable_x"

DELIMITER_PARENTAL_CONTROL_ITEM = "&#62"


### ENDPOINTS
ENDPOINT_LOGIN = "login.cgi"
ENDPOINT_LOGOUT = "Logout.asp"

ENDPOINT = {
    APPLY: "apply.cgi",
    COMMAND: "applyapp.cgi",
    DEVICEMAP: "ajax_status.xml",
    ETHERNET_PORTS: "ajax_ethernet_ports.asp",
    FIRMWARE: "detect_firmware.asp",
    HOOK: "appGet.cgi",
    LEDG: "set_ledg.cgi",
    # NETWORKMAPD: "update_networkmapd.asp",
    ONBOARDING: "ajax_onboarding.asp",
    PORT_STATUS: "get_port_status.cgi",
    RGB: "light_effect.html",
    STATE: "state.js",
    SYSINFO: "ajax_sysinfo.asp",
    TEMPERATURE: "ajax_coretmp.asp",
    UPDATE_CLIENTS: "update_clients.asp",
    VPN: "ajax_vpn_status.asp",
}

ENDHOOKS: dict[str, Any] = {
    DEVICES: [("get_clientlist", "")],
    LIGHT: [
        (NVRAM_GET, LED_VAL),
    ],
    MAIN: [
        (CPU_USAGE, APPOBJ),
        (MEMORY_USAGE, APPOBJ),
        (NETDEV, APPOBJ),
        (WANLINK_STATE, APPOBJ),
    ],
    NVRAM: None,
    PARENTAL_CONTROL: [
        (NVRAM_GET, KEY_PARENTAL_CONTROL_MAC),
        (NVRAM_GET, KEY_PARENTAL_CONTROL_NAME),
        (NVRAM_GET, KEY_PARENTAL_CONTROL_STATE),
        (NVRAM_GET, KEY_PARENTAL_CONTROL_TIMEMAP),
        (NVRAM_GET, KEY_PARENTAL_CONTROL_TYPE),
    ],
    PORT_FORWARDING: [
        (NVRAM_GET, KEY_PORT_FORWARDING_LIST),
        (NVRAM_GET, KEY_PORT_FORWARDING_STATE),
    ],
}
ENDPOINT_ARGS = {
    PORT_STATUS: {
        MAC: f"{NODE}_{MAC}",
    },
}

### History-dependent values in the monitor to be removed to prevent errors
HD_DATA: tuple[tuple[str, ...], ...] = (
    (MAIN, CPU),
    (MAIN, NETWORK),
)

### REQUIREMENTS
# Don't create loops
CONST_REQUIRE_MONITOR = {
    WLAN: MAIN,
}
MONITOR_REQUIRE_CONST = {
    NVRAM: WLAN,
}

### ASUS NUMERICS, RANGES & MAPS

# Map of the parameters for connected device
# value_to_find: [where_to_search, method to convert]
# List is sorted by importance with the most important first
MAP_CONNECTED_DEVICE: dict[str, list[SearchKey]] = {
    "connected_since": [
        SearchKey("wlConnectTime", time_from_delta),
    ],
    CONNECTION_TYPE: [
        SearchKey(CONNECTION_TYPE),
        SearchKey("isWL", int_from_str),
    ],
    GUEST: [
        SearchKey(GUEST),
        SearchKey("isGN", int_from_str),
    ],
    "internet_mode": [
        SearchKey("internetMode"),
    ],
    "internet_state": [
        SearchKey("internetState", bool_from_any),
    ],
    IP: [
        SearchKey(IP),
    ],
    "ip_method": [
        SearchKey("ipMethod"),
    ],
    MAC: [
        SearchKey(MAC),
    ],
    NAME: [
        SearchKey("nickName"),
        SearchKey(NAME),
        SearchKey(MAC),
    ],
    NODE: [
        SearchKey(NODE),
    ],
    ONLINE: [
        SearchKey(ONLINE),
        SearchKey("isOnline", bool_from_any),
    ],
    RSSI: [
        SearchKey(RSSI, int_from_str),
    ],
    "rx_speed": [
        SearchKey("curRx", float_from_str),
    ],
    "tx_speed": [
        SearchKey("curTx", float_from_str),
    ],
}
MAP_CPU: tuple[Key, ...] = (Key(TOTAL), Key(USAGE, USED))
MAP_IDENTITY: tuple[Key, ...] = (
    Key("serial_no", "serial"),
    Key("label_mac", MAC),
    Key("lan_hwaddr", "lan_mac"),
    Key("wan_hwaddr", "wan_mac"),
    Key("productid", "model"),
    Key("firmver", "fw_major"),
    Key("buildno", "fw_minor"),
    Key("extendno", "fw_build"),
    Key("rc_support", "services", method=service_support),
    Key("ss_support", "services", method=service_support),
    Key(LED_VAL, LED, method=exists_or_not),
)
MAP_NETWORK: tuple[Key, ...] = (
    Key("INTERNET", WAN),
    Key("INTERNET1", USB),
    Key("WIRED", WIRED),
    Key("BRIDGE", BRIDGE),
    Key("WIRELESS0", WLAN_2GHZ),
    Key("WIRELESS1", WLAN_5GHZ),
    Key("WIRELESS2", WLAN_5GHZ2),
    Key("WIRELESS3", WLAN_6GHZ),
    Key("LACP1", f"{LACP}1"),
    Key("LACP2", f"{LACP}2"),
)
MAP_NVRAM = {
    WLAN: [
        SearchKey("wl{}_auth_mode_x"),
        SearchKey("wl{}_bw"),
        SearchKey("wl{}_channel"),
        SearchKey("wl{}_chanspec"),
        SearchKey("wl{}_closed", bool_from_any),
        SearchKey("wl{}_country_code"),
        SearchKey("wl{}_crypto"),
        SearchKey("wl{}_gmode_check"),
        SearchKey("wl{}_maclist_x"),
        SearchKey("wl{}_macmode"),
        SearchKey("wl{}_mbo_enable"),
        SearchKey("wl{}_mfp"),
        SearchKey("wl{}_nmode_x"),
        SearchKey("wl{}_optimizexbox_ckb"),
        SearchKey("wl{}_radio", bool_from_any),
        SearchKey("wl{}_radius_ipaddr"),
        SearchKey("wl{}_radius_key"),
        SearchKey("wl{}_radius_port"),
        SearchKey("wl{}_ssid"),
        SearchKey("wl{}_wpa_gtk_rekey"),
        SearchKey("wl{}_wpa_psk"),
    ],
    GWLAN: [
        SearchKey("wl{}_akm"),
        SearchKey("wl{}_ap_isolate"),
        SearchKey("wl{}_auth"),
        SearchKey("wl{}_auth_mode"),
        SearchKey("wl{}_auth_mode_x"),
        SearchKey("wl{}_bridge"),
        SearchKey("wl{}_bss_enabled", bool_from_any),
        SearchKey("wl{}_bss_maxassoc", int_from_str),
        SearchKey("wl{}_bw_dl", int_from_str),  # Bandwidth limit download
        SearchKey("wl{}_bw_enabled", bool_from_any),  # Bandwidth limit switch
        SearchKey("wl{}_bw_ul", int_from_str),  # Bandwidth limit upload
        SearchKey("wl{}_closed", bool_from_any),
        SearchKey("wl{}_crypto"),
        SearchKey("wl{}_expire", int_from_str),  # Expire time in s
        SearchKey("wl{}_expire_tmp", int_from_str),  # Expire time left in s
        SearchKey("wl{}_gn_wbl_enable"),
        SearchKey("wl{}_gn_wbl_rule"),
        SearchKey("wl{}_gn_wbl_type"),
        SearchKey("wl{}_hwaddr"),  # MAC address
        SearchKey("wl{}_ifname"),  # Interface name
        SearchKey("wl{}_infra"),
        SearchKey("wl{}_key"),
        SearchKey("wl{}_key1"),
        SearchKey("wl{}_key2"),
        SearchKey("wl{}_key3"),
        SearchKey("wl{}_key4"),
        SearchKey("wl{}_lanaccess", bool_from_any),  # LAN access
        SearchKey("wl{}_maclist"),
        SearchKey("wl{}_macmode"),
        SearchKey("wl{}_maxassoc", int_from_str),
        SearchKey("wl{}_mbss"),
        SearchKey("wl{}_mfp"),
        SearchKey("wl{}_mode"),
        SearchKey("wl{}_net_reauth", int_from_str),
        SearchKey("wl{}_preauth"),
        SearchKey("wl{}_radio", bool_from_any),
        SearchKey("wl{}_radius_ipaddr"),
        SearchKey("wl{}_radius_key"),
        SearchKey("wl{}_radius_port", int_from_str),
        SearchKey("wl{}_sae_anti_clog_threshold"),
        SearchKey("wl{}_sae_groups"),
        SearchKey("wl{}_sae_sync"),
        SearchKey("wl{}_ssid"),  # SSID
        SearchKey("wl{}_sta_retry_time"),
        SearchKey("wl{}_sync_node", bool_from_any),  # Sync AiMesh nodes
        SearchKey("wl{}_unit"),  # GWLAN unit id
        SearchKey("wl{}_wep", bool_from_any),
        SearchKey("wl{}_wep_x", bool_from_any),
        SearchKey("wl{}_wfi_enable", bool_from_any),
        SearchKey("wl{}_wfi_pinmode"),
        SearchKey("wl{}_wme"),
        SearchKey("wl{}_wme_bss_disable", bool_from_any),
        SearchKey("wl{}_wpa_gtk_rekey"),
        SearchKey("wl{}_wpa_psk"),  # Password
        SearchKey("wl{}_wps_mode"),
    ],
}
MAP_OVPN_CLIENT = (
    Key("REMOTE", "remote", method=ovpn_remote_fom_str),
    Key("Updated", "datetime", method=datetime_from_str),
    Key("TUN/TAP read bytes", "tun_tap_read", method=int_from_str),
    Key("TUN/TAP write bytes", "tun_tap_write", method=int_from_str),
    Key("TCP/UDP read bytes", "tcp_udp_read", method=int_from_str),
    Key("TCP/UDP write bytes", "tcp_udp_write", method=int_from_str),
    Key("Auth read bytes", "auth_read", method=int_from_str),
    Key("pre-compress bytes", "pre_compress", method=int_from_str),
    Key("post-compress bytes", "post_compress", method=int_from_str),
    Key("pre-decompress bytes", "pre_decompress", method=int_from_str),
    Key("post-decompress bytes", "post_decompress", method=int_from_str),
)
MAP_OVPN_STATUS = {
    -1: "error",
    0: "disconnected",
    1: "connecting",
    2: "connected",
}
MAP_OVPN_SERVER = (
    Key("CLIENT_LIST", "client_list"),
    Key("ROUTING_TABLE", "routing_table"),
)
MAP_PARENTAL_CONTROL_ITEM: tuple[Key, ...] = (
    Key(KEY_PARENTAL_CONTROL_MAC, MAC),
    Key(KEY_PARENTAL_CONTROL_NAME, NAME),
    Key(KEY_PARENTAL_CONTROL_TIMEMAP, TIMEMAP),
    Key(KEY_PARENTAL_CONTROL_TYPE, TYPE, int_from_str),
)
MAP_PARENTAL_CONTROL_TYPE = {
    0: DISABLE,
    1: TIME,
    2: BLOCK,
}
MAP_PORTS: tuple[Key, ...] = ()
MAP_RAM: tuple[Key, ...] = (Key(FREE), Key(TOTAL), Key(USED))
MAP_WAN = (
    Key("wanstate", STATE),
    Key("wansbstate", "bstate"),
    Key("wanauxstate", "aux"),
    Key("autodet_state"),
    Key("autodet_auxstate"),
    Key("wanlink_status", STATUS, method=bool_from_any),
    Key("wanlink_type", "ip_type"),
    Key("wanlink_ipaddr", IP),
    Key("wanlink_netmask", "mask"),
    Key("wanlink_gateway", "gateway"),
    Key("wanlink_dns", "dns"),
    Key("wanlink_lease", "lease"),
    Key("wanlink_expires", "expires"),
    Key("is_private_subnet", "private_subnet", method=int_from_str),
    Key("wanlink_xtype", "xtype"),
    Key("wanlink_xipaddr", "xip"),
    Key("wanlink_xnetmask", "xmask"),
    Key("wanlink_xgateway", "xgateway"),
    Key("wanlink_xdns", "xdns"),
    Key("wanlink_xlease", "xlease"),
    Key("wanlink_xexpires", "xexpires"),
)
RANGE_CPU_CORES = range(1, 9)  # 8 cores from 1 to 8
RANGE_GWLAN = range(1, 4)  # 3 guest WLANs from 1 to 3 per each WLAN
RANGE_OVPN_CLIENTS = range(1, 6)  # 5 Open VPN clients from 1 to 5
RANGE_OVPN_SERVERS = range(1, 3)  # 2 Open VPN servers from 1 to 2
RANGE_WLAN = range(0, 4)  # 4 WLANs from 0 to 3

### TYPES

PORT_TYPES = {
    "L": LAN,
    "U": USB,
    "W": WAN,
}

SPEED_TYPES = {
    "X": 0,
    "M": 100,
    "G": 1000,
    "Q": 2500,
}

TRAFFIC_TYPE: tuple[str, ...] = (
    RX,
    TX,
)

WLAN_TYPE: dict[str, str] = {
    WLAN_2GHZ: 0,
    WLAN_5GHZ: 1,
    WLAN_5GHZ2: 2,
    WLAN_6GHZ: 3,
}

### CONVERTERS

CONVERTERS = {
    PORT_STATUS: [
        Key("is_on", STATE, method=bool_from_any),
        Key("max_rate", method=int_from_str),
        Key(LINK_RATE, method=int_from_str),
    ]
}

### CONSTANTS, DATA TYPES AND COMMON PARAMETERS
CONST_BITSINBYTE = 8
CONST_PERCENTS = 100
CONST_ZERO = 0.0


# CHECK LEGACY FROM HERE -->

DATA_FREE = "free"
DATA_TOTAL = "total"
DATA_USAGE = "usage"
DATA_USED = "used"
DATA_RX = "rx"
DATA_TX = "tx"
DATA_TRAFFIC = [DATA_RX, DATA_TX]
DATA_BY_CORE = "core_{}"
DATA_ADD_SPEED = "{}_speed"

DEFAULT_CACHE_TIME = 5
DEFAULT_PORT = {
    "http": 80,
    "https": 8443,
}
DEFAULT_SLEEP_RECONNECT = 5
DEFAULT_SLEEP_CONNECTING = 1
DEFAULT_SLEEP_TIME = 0.1
DEFAULT_TRAFFIC_OVERFLOW = 4294967296
DEFAULT_USAGE_DIGITS = 2

PARAM_COLOR = "color"
PARAM_COUNT = "count"
PARAM_ERRNO = "errno"
PARAM_IP = "ip"
PARAM_MODE = "mode"
PARAM_RIP = "rip"
PARAM_STATE = "state"
PARAM_STATUS = "status"
PARAM_UNKNOWN = "unknown"

# OUR REGEX
REGEX_VARIABLES = r'([a-zA-Z0-9\_-]+)\s*=\s*"(.*?)"(?=;)'

# VALUES TO IGNORE
VALUES_TO_IGNORE = [str(), "None", "0.0.0.0"]

# Services to recover LED state
TRACK_SERVICES_LED = [
    "restart_wireless",
]


### ASUSWRT KEYS, MAPS AND VALUES
AR_DEFAULT_LEDG = 8

AR_KEY_AURARGB = "aurargb"
AR_KEY_HEADER = "HEADER"
AR_KEY_LED = "led_val"
AR_KEY_LEDG_COUNT = "ledg_count"
AR_KEY_LEDG_SCHEME = "ledg_scheme"
AR_KEY_LEDG_SCHEME_OLD = "ledg_scheme_old"
AR_KEY_LEDG_RGB = "ledg_rgb{}"
AR_KEY_SERVICE_COMMAND = "rc_service"
AR_KEY_SERVICE_MODIFY = "modify"
AR_KEY_SERVICE_REPLY = "run_service"

AR_LEDG_MODE: dict[int, str] = {
    1: "Gradient",
    2: "Static",
    3: "Breathing",
    4: "Evolution",
    5: "Rainbow",
    6: "Wave",
    7: "Marquee",
}

AR_MAP_RGB: dict[int, str] = {
    0: "r",
    1: "g",
    2: "b",
}
AR_MAP_SYSINFO: dict[str, list[Key]] = {
    "conn_stats_arr": [
        Key(0, "conn_total", method=int_from_str),
        Key(1, "conn_active", method=int_from_str),
    ],
    "cpu_stats_arr": [
        Key(0, "load_avg_1", method=float_from_str),
        Key(1, "load_avg_5", method=float_from_str),
        Key(2, "load_avg_15", method=float_from_str),
    ],
    "mem_stats_arr": [
        Key(0, "ram_total", method=float_from_str),
        Key(1, "ram_free", method=float_from_str),
        Key(2, "buffers", method=float_from_str),
        Key(3, "cache", method=float_from_str),
        Key(4, "swap_size", method=float_from_str),
        Key(5, "swap_total", method=float_from_str),
        Key(6, "nvram_used", method=int_from_str),
        Key(7, "jffs"),
    ],
    "wlc_24_arr": [
        Key(0, "2ghz_clients_associated", method=bool_or_int),
        Key(1, "2ghz_clients_authorized", method=bool_or_int),
        Key(2, "2ghz_clients_authenticated", method=bool_or_int),
    ],
    "wlc_51_arr": [
        Key(0, "5ghz_clients_associated", method=bool_or_int),
        Key(1, "5ghz_clients_authorized", method=bool_or_int),
        Key(2, "5ghz_clients_authenticated", method=bool_or_int),
    ],
    "wlc_52_arr": [
        Key(0, "5ghz2_clients_associated", method=bool_or_int),
        Key(1, "5ghz2_clients_authorized", method=bool_or_int),
        Key(2, "5ghz2_clients_authenticated", method=bool_or_int),
    ],
}
AR_MAP_TEMPERATURE: dict[str, list[str]] = {
    "2ghz": [
        'curr_coreTmp_2_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_0_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_wl0_raw="([0-9.]+)&deg;C',
    ],
    "5ghz": [
        'curr_coreTmp_5_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_1_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_wl1_raw="([0-9.]+)&deg;C',
    ],
    "5ghz2": [
        'curr_coreTmp_52_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_2_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_wl2_raw="([0-9.]+)&deg;C',
    ],
    "6ghz": [
        'curr_coreTmp_3_raw="([0-9.]+)&deg;C',
        'curr_coreTmp_wl3_raw="([0-9.]+)&deg;C',
    ],
    "cpu": ['curr_cpuTemp="([0-9.]+)"', 'curr_coreTmp_cpu="([0-9.]+)"'],
}

### ASUSWRT SERVICES

AR_FIRMWARE_CHECK_COMMAND = "firmware_check"
AR_SERVICE_COMMAND: dict[str, str] = {
    "force_restart": "restart_{}_force",
    "start": "start_{}",
    "stop": "stop_{}",
    "restart": "restart_{}",
}
AR_SERVICE_CONTROL: dict[str, list[str]] = {
    "firewall": [
        "restart",
    ],
    "ftpd": [
        "force_restart",
        "start",
        "stop",
        "restart",
    ],
    "httpd": [
        "restart",
    ],
    "wireless": [
        "restart",
    ],
}
AR_SERVICE_DROP_CONNECTION: list[str] = {
    "httpd",
}

### ERRORS AND MESSAGES
ERROR_IDENTITY = "Cannot obtain identity from the device {}. Exception summary{}"
ERROR_PARSING = (
    "Failed parsing value '{}'. Please report this issue. Exception summary: {}"
)
ERROR_SERVICE = (
    "Error calling service '{}'. "
    "Service did not return any expected value in the reply: {}"
)
ERROR_SERVICE_UNKNOWN = "Unknown service '{}' with mode '{}'"
ERROR_VALUE = "Wrong value '{}' with original exception: {}"
ERROR_VALUE_TYPE = "Wrong value '{}' of type {}"
ERROR_ZERO_DIVISION = "Zero division allert: {}"

INTERFACE_TYPE = {
    "wan_ifnames": "wan",
    "wl_ifnames": "wlan",
    "wl0_vifnames": "gwlan2",
    "wl1_vifnames": "gwlan5",
    "lan_ifnames": "lan",
}

NVRAM_LIST = {
    "DEVICE": [
        "productid",
        "serial_no",
        "label_mac",
        "model",
        "pci/1/1/ATE_Brand",
    ],
    "MODE": [
        "sw_mode",
        "wlc_psta",
        "wlc_express",
    ],
    "FIRMWARE": [
        "firmver",
        "buildno",
        "extendno",
    ],
    "FUNCTIONS": [
        "rc_support",
        "ss_support",
    ],
    "INTERFACES": [
        "wans_dualwan",
        "wan_ifnames",
        "wl_ifnames",
        "wl0_vifnames",
        "wl1_vifnames",
        "lan_ifnames",
        "br0_ifnames",
    ],
    "REBOOT": [
        "reboot_schedule",
        "reboot_schedule_enable",
        "reboot_time",
    ],
    "WAN0": [
        "link_wan",
        "wan0_auxstate_t",
        "wan0_dns",
        "wan0_enable",
        "wan0_expires",
        "wan0_gateway",
        "wan0_gw_ifname",
        "wan0_gw_mac",
        "wan0_hwaddr",
        "wan0_ipaddr",
        "wan0_is_usb_modem_ready",
        "wan0_primary",
        "wan0_proto",
        "wan0_realip_ip",
        "wan0_realip_state",
        "wan0_sbstate_t",
        "wan0_state_t",
        "wan0_upnp_enable",
    ],
    "WAN1": [
        "link_wan1",
        "wan1_auxstate_t",
        "wan1_dns",
        "wan1_enable",
        "wan1_gateway",
        "wan1_gw_ifname",
        "wan1_gw_mac",
        "wan1_hwaddr",
        "wan1_ipaddr",
        "wan1_is_usb_modem_ready",
        "wan1_primary",
        "wan1_proto",
        "wan1_realip_ip",
        "wan1_realip_state",
        "wan1_sbstate_t",
        "wan1_state_t",
        "wan1_upnp_enable",
    ],
    "LAN": [
        "lan_gateway",
        "lan_hwaddr",
        "lan_ifname",
        "lan_ifnames",
        "lan_ipaddr",
        "lan_netmask",
        "lan_proto",
        "lanports",
    ],
    "WLAN0": [
        "wl0_radio",
        "wl0_ssid",
        "wl0_chanspec",
        "wl0_closed",
        "wl0_nmode_x",
        "wl0_bw",
        "wl0_auth_mode_x",
        "wl0_crypto",
        "wl0_wpa_psk",
        "wl0_mfp",
        "wl0_mbo_enable",
        "wl0_country_code",
        "wl0_maclist_x",
        "wl0_macmode",
    ],
    "WLAN1": [
        "wl1_radio",
        "wl1_ssid",
        "wl1_chanspec",
        "wl1_closed",
        "wl1_nmode_x",
        "wl1_bw",
        "wl1_auth_mode_x",
        "wl1_crypto",
        "wl1_wpa_psk",
        "wl1_mfp",
        "wl1_mbo_enable",
        "wl1_country_code",
        "wl1_maclist_x",
        "wl1_macmode",
    ],
    "GWLAN01": [
        "wl0.1_bss_enabled",
        "wl0.1_ssid",
        "wl0.1_auth_mode_x",
        "wl0.1_crypto",
        "wl0.1_key",
        "wl0.1_wpa_psk",
        "wl0.1_lanaccess",
        "wl0.1_expire",
        "wl0.1_expire_tmp",
        "wl0.1_macmode",
        "wl0.1_mbss",
        "wl0.1_sync_mode",
    ],
    "GWLAN02": [
        "wl0.2_bss_enabled",
        "wl0.2_ssid",
        "wl0.2_auth_mode_x",
        "wl0.2_crypto",
        "wl0.2_key",
        "wl0.2_wpa_psk",
        "wl0.2_lanaccess",
        "wl0.2_expire",
        "wl0.2_expire_tmp",
        "wl0.2_macmode",
        "wl0.2_mbss",
        "wl0.2_sync_mode",
    ],
    "GWLAN03": [
        "wl0.3_bss_enabled",
        "wl0.3_ssid",
        "wl0.3_auth_mode_x",
        "wl0.3_crypto",
        "wl0.3_key",
        "wl0.3_wpa_psk",
        "wl0.3_lanaccess",
        "wl0.3_expire",
        "wl0.3_expire_tmp",
        "wl0.3_macmode",
        "wl0.3_mbss",
        "wl0.3_sync_mode",
    ],
    "GWLAN11": [
        "wl1.1_bss_enabled",
        "wl1.1_ssid",
        "wl1.1_auth_mode_x",
        "wl1.1_crypto",
        "wl1.1_key",
        "wl1.1_wpa_psk",
        "wl1.1_lanaccess",
        "wl1.1_expire",
        "wl1.1_expire_tmp",
        "wl1.1_macmode",
        "wl1.1_mbss",
        "wl1.1_sync_mode",
    ],
    "GWLAN12": [
        "wl1.2_bss_enabled",
        "wl1.2_ssid",
        "wl1.2_auth_mode_x",
        "wl1.2_crypto",
        "wl1.2_key",
        "wl1.2_wpa_psk",
        "wl1.2_lanaccess",
        "wl1.2_expire",
        "wl1.2_expire_tmp",
        "wl1.2_macmode",
        "wl1.2_mbss",
        "wl1.2_sync_mode",
    ],
    "GWLAN13": [
        "wl1.3_bss_enabled",
        "wl1.3_ssid",
        "wl1.3_auth_mode_x",
        "wl1.3_crypto",
        "wl1.3_key",
        "wl1.3_wpa_psk",
        "wl1.3_lanaccess",
        "wl1.3_expire",
        "wl1.3_expire_tmp",
        "wl1.3_macmode",
        "wl1.3_mbss",
        "wl1.3_sync_mode",
    ],
}

KEY_NVRAM_GET = "nvram_get"
KEY_ACTION_MODE = "action_mode"
KEY_HOOK = "hook"
DEFAULT_ACTION_MODE = "apply"


### DEVICEMAP

# These values are just stored directly in this order in the corresponding node
DEVICEMAP_BY_INDEX = {
    WAN: {
        "wan": [
            "status",
            "sbstatus",
            "auxstatus",
        ],
    },
    f"{WAN}0": {
        "first_wan": [
            "status",
            "sbstatus",
            "auxstatus",
        ],
    },
    f"{WAN}1": {
        "second_wan": [
            "status",
            "sbstatus",
            "auxstatus",
        ],
    },
    USB: {
        "usb": [
            "status",
        ],
    },
}


# Not implemented yet
# psta:wlc_state=0;wlc_state_auth=0;
DEVICEMAP_SPECIAL = {
    WAN: {
        "psta": [
            "wlc_state",
            "wlc_state_auth",
        ],
    },
}

# These values are stored as "key=value"
DEVICEMAP_GENERAL = {
    WAN: {
        "wan": [
            "monoClient",
            "wlc_state",
            "wlc_sbstate",
            "wifi_hw_switch",
            "ddnsRet",
            "ddnsUpdate",
            "wan_line_state",
            "wlan0_radio_flag",
            "wlan1_radio_flag",
            "wlan2_radio_flag",
            "data_rate_info_2g",
            "data_rate_info_5g",
            "data_rate_info_5g_2",
            "wan_diag_state",
            "active_wan_unit",
            "wlc0_state",
            "wlc1_state",
            "rssi_2g",
            "rssi_5g",
            "rssi_5g_2",
            "link_internet",
            "wlc2_state",
            "le_restart_httpd",
        ],
    },
    VPN: {
        "vpn": [
            "vpnc_proto",
            "vpnc_state_t",
            "vpnc_sbstate_t",
            "vpn_client1_state",
            "vpn_client2_state",
            "vpn_client3_state",
            "vpn_client4_state",
            "vpn_client5_state",
            "vpnd_state",
            "vpn_client1_errno",
            "vpn_client2_errno",
            "vpn_client3_errno",
            "vpn_client4_errno",
            "vpn_client5_errno",
        ],
    },
    SYS: {
        "sys": [
            "uptimeStr",
        ],
    },
    QTN: {
        "qtn": [
            "qtn_state",
        ],
    },
    USB: {
        "usb": [
            "modem_enable",
        ],
    },
    f"{WAN}0": {
        "wan": [
            "wan0_enable",
            "wan0_realip_state",
            "wan0_ipaddr",
            "wan0_realip_ip",
        ],
    },
    f"{WAN}1": {
        "wan": [
            "wan1_enable",
            "wan1_realip_state",
            "wan1_ipaddr",
            "wan1_realip_ip",
        ],
    },
    SIM: {
        "sim": [
            "sim_state",
            "sim_signal",
            "sim_operation",
            "sim_isp",
            "roaming",
            "roaming_imsi",
            "sim_imsi",
            "g3err_pin",
            "pin_remaining_count",
            "modem_act_provider",
            "rx_bytes",
            "tx_bytes",
            "modem_sim_order",
        ],
    },
    DHCP: {
        "dhcp": [
            "dnsqmode",
        ],
    },
    DIAG: {
        "diag": [
            "diag_dblog_enable",
            "diag_dblog_remaining",
        ]
    },
}

# This should be cleared from the data
DEVICEMAP_CLEAR = {
    WAN: {
        "data_rate_info_2g": '"',
        "data_rate_info_5g": '"',
        "data_rate_info_5g_2": '"',
    }
}



"""Parsers module for AsusRouter"""

from __future__ import annotations

import json
import logging
import re
from datetime import datetime, timedelta
from typing import Any

import xmltodict
from dateutil.parser import parse as dtparse

from asusrouter.const import (
    AR_DEFAULT_LEDG,
    AR_KEY_HEADER,
    AR_MAP_RGB,
    AR_MAP_SYSINFO,
    AR_MAP_TEMPERATURE,
    CONST_BITSINBYTE,
    CONST_ZERO,
    CPU,
    DATA_ADD_SPEED,
    DEVICEMAP,
    DEVICEMAP_BY_INDEX,
    DEVICEMAP_CLEAR,
    DEVICEMAP_GENERAL,
    ENDPOINT,
    ERROR_VALUE,
    ERROR_VALUE_TYPE,
    FIRMWARE,
    IP,
    MAP_CPU,
    MAP_NETWORK,
    MAP_OVPN_CLIENT,
    MAP_OVPN_SERVER,
    MAP_RAM,
    MAP_WAN,
    MEM,
    ONBOARDING,
    PARAM_IP,
    PARAM_RIP,
    PARAM_STATE,
    RANGE_CPU_CORES,
    RANGE_OVPN_CLIENTS,
    RANGE_OVPN_SERVERS,
    REGEX_VARIABLES,
    RIP,
    STATE,
    STATUS,
    SYSINFO,
    TOTAL,
    TRAFFIC_TYPE,
    UPDATE_CLIENTS,
    VALUES_TO_IGNORE,
    VPN,
    VPN_CLIENT,
    VPN_SERVER,
)
from asusrouter.dataclass import AiMeshDevice, Firmware
from asusrouter.error import AsusRouterDataProcessError, AsusRouterValueError
from asusrouter.util import calculators, converters

_LOGGER = logging.getLogger(__name__)


def cpu_usage(raw: dict[str, Any]) -> dict[str, Any]:
    """CPU usage parser"""

    cpu = {}

    # Populate total
    cpu[TOTAL] = {}
    for item in MAP_CPU:
        cpu[TOTAL][item.get()] = CONST_ZERO

    # Data / core
    for core in RANGE_CPU_CORES:
        if not f"{CPU}{core}_{TOTAL}" in raw:
            break

        cpu[core] = {}

        for item in MAP_CPU:
            key = f"{CPU}{core}_{item}"
            new_key = item.get()
            if key in raw:
                try:
                    cpu[core][new_key] = int(raw[key])
                except ValueError as ex:
                    raise (
                        AsusRouterValueError(ERROR_VALUE.format(raw[key], str(ex)))
                    ) from ex
                # Add this to total as well
                cpu[TOTAL][new_key] += cpu[core][new_key]

    return cpu


def ram_usage(raw: dict[str, Any]) -> dict[str, Any]:
    """RAM usage parser"""

    ram = {}

    for item in MAP_RAM:
        if f"{MEM}_{item.value}" in raw:
            try:
                ram[item.get()] = int(raw[f"{MEM}_{item.value}"])
            except ValueError as ex:
                raise (
                    AsusRouterValueError(
                        ERROR_VALUE.format(raw[f"{MEM}_{item.value}"], str(ex))
                    )
                ) from ex

    return ram


def network_usage(raw: dict[str, Any]) -> dict[str, Any]:
    """Network usage parser"""

    network = {}
    for interface in MAP_NETWORK:
        data = {}
        for traffic in TRAFFIC_TYPE:
            try:
                value = converters.int_from_str(
                    raw.get(f"{interface.value}_{traffic}"), base=16
                )
                if value:
                    data[traffic] = value
            except Exception:
                continue
        if len(data) > 0:
            network[interface.get()] = data

    return network


def network_speed(
    after: dict[str, dict[str, float]],
    before: dict[str, dict[str, float]],
    time_delta: float,
) -> dict[str, dict[str, float]]:
    """
    Network speed calculator

    Parameters
    -----
    `after`: current values. Outer dictionary `(groups)` contains inner dictionary `(types)`.
    On each `(type)` calculations are performed

    `before`: previous values

    `time_delta`: time between measurements

    Returns
    -----
    `after` with speeds append to `(types)`
    """

    for group in after:
        if group in before:
            speed = {}
            for type in after[group]:
                if type in before[group]:
                    speed[
                        DATA_ADD_SPEED.format(type)
                    ] = CONST_BITSINBYTE * calculators.speed(
                        after=after[group][type],
                        before=before[group][type],
                        time_delta=time_delta,
                    )
                else:
                    speed[DATA_ADD_SPEED.format(type)] = CONST_ZERO
            after[group] |= speed

    return after


def wan_state(raw: dict[str, Any]) -> dict[str, Any]:
    """WAN status parser"""

    values = {}

    for key in MAP_WAN:
        if key.value in raw and raw[key.value] != str():
            try:
                values[key.get()] = (
                    key.method(raw[key.value]) if key.method else raw[key.value]
                )
            except AsusRouterValueError as ex:
                _LOGGER.warning(
                    "Failed parsing value '%s'. Please report this issue. Exception summary: %s",
                    key.value,
                    str(ex),
                )

    return values


def uptime(data: str) -> datetime:
    """Uptime -> boot time parser"""

    try:
        part = data.split("(")
        seconds = int(re.search("([0-9]+)", part[1]).group())
        when = dtparse(part[0])
        boot = when - timedelta(seconds=seconds)
    except ValueError as ex:
        raise (AsusRouterValueError(ERROR_VALUE.format(data, str(ex)))) from ex

    return boot


def devicemap(devicemap: dict[str, Any]) -> dict[str, Any]:
    """Devicemap parser"""

    data = {}

    # Get values only with index
    for node, body in DEVICEMAP_BY_INDEX.items():
        _node = {}
        for key in body:
            for el in range(len(DEVICEMAP_BY_INDEX[node][key])):
                _node[DEVICEMAP_BY_INDEX[node][key][el]] = devicemap[key][el]
            del devicemap[key][0 : (len(DEVICEMAP_BY_INDEX[node][key]) - 1)]
        data[node] = _node

    # Get values by key
    for node, body in DEVICEMAP_GENERAL.items():
        _node = {}
        for key in body:
            for el in DEVICEMAP_GENERAL[node][key]:
                if key in devicemap:
                    if isinstance(devicemap[key], str):
                        if el in devicemap[key]:
                            _node[el] = devicemap[key].replace(f"{el}=", "")
                            break
                    else:
                        for value in devicemap[key]:
                            if el in value:
                                _node[el] = value.replace(f"{el}=", "")
                                break
        if node in data:
            data[node].update(_node)
        else:
            data[node] = _node

    # Clear values from useless symbols
    for node, body in DEVICEMAP_CLEAR.items():
        for value in body:
            data[node][value] = data[node][value].replace(
                DEVICEMAP_CLEAR[node][value], ""
            )

    return data


def ledg_count(raw: str) -> int:
    """LEDG count parser"""

    if not isinstance(raw, str):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    value = re.search("ledg_count: ([0-9]+)", raw)
    if value:
        count = int(value[1])
    else:
        count = 0

    return count


def temperatures(raw: str) -> dict[str, Any]:
    """Temperature parser"""

    if not isinstance(raw, str):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    temp = {}

    for sensor, map in AR_MAP_TEMPERATURE.items():
        for reg in map:
            value = re.search(reg, raw)
            if value:
                temp[sensor] = float(value[1])

    return temp


def rgb(raw: str, num: int = AR_DEFAULT_LEDG) -> dict[int, dict[str, int]]:
    """RGB to channels parser"""

    if not isinstance(raw, str):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    leds = {}

    color = re.findall("([0-9]+)", raw)
    length = len(color)

    for led in range(1, num):
        ind = (led - 1) * 3
        if ind + 2 > length:
            break
        if not led in leds:
            leds[led] = {}
        for channel, code in AR_MAP_RGB.items():
            leds[led][code] = color[ind + channel]

    return leds


def sysinfo(raw: str) -> dict[str, Any]:
    """Sysinfo parser"""

    raw = raw.replace("=", '":')
    raw = raw.replace(";", ',"')
    raw = '{"' + raw[:-2] + "}"

    try:
        data = json.loads(raw)
    except json.JSONDecodeError as ex:
        raise AsusRouterDataProcessError from ex

    result = {}

    for set, keys in AR_MAP_SYSINFO.items():
        if set in data:
            for key in keys:
                try:
                    result[key.get()] = (
                        key.method(data[set][key.value])
                        if key.method
                        else data[set][key.value]
                    )
                except AsusRouterValueError as ex:
                    _LOGGER.warning(
                        "Failed parsing value '%s'. Please report this issue. Exception summary: %s",
                        key.value,
                        str(ex),
                    )

    return result


def onboarding(raw: str) -> dict[str, Any]:
    """Onboarding parser"""

    raw = raw.replace("\ufeff", "")
    raw = raw.replace("=", '":')
    raw = raw.replace(";", ',"')
    raw = raw.replace("[0]", "")
    raw = '{"' + raw[:-2] + "}"

    try:
        data = json.loads(raw)
    except json.JSONDecodeError as ex:
        raise AsusRouterDataProcessError from ex

    return data


def aimesh_node(raw: dict[str, Any]) -> AiMeshDevice:
    """AiMesh node parser"""

    AP = {
        "2g": "2ghz",
        "5g": "5ghz",
        "5g1": "5ghz2",
        "6g": "6ghz",
        "dwb": "dwb",
    }

    FREQUENCIES = [
        "2g",
        "5g",
        "6g",
    ]

    ap = {}
    for el, value in AP.items():
        if f"ap{el}" in raw and raw[f"ap{el}"] is not str():
            ap[value] = raw[f"ap{el}"]

    parent = {}
    for el in FREQUENCIES:
        if f"pap{el}" in raw and raw[f"pap{el}"] is not str():
            parent["connection"] = AP[el]
            parent["mac"] = raw[f"pap{el}"]
            parent["rssi"] = converters.none_or_str(raw.get(f"rssi{el}"))
            parent["ssid"] = converters.none_or_str(raw.get(f"pap{el}_ssid"))

    level = converters.int_from_str(raw.get("level", "0"))
    state = "router" if level == 0 else "node"

    return AiMeshDevice(
        status=converters.bool_from_any(raw.get("online", 0)),
        alias=raw.get("alias", None),
        model=raw.get("ui_model_name", raw.get("model_name", None)),
        product_id=raw.get("product_id"),
        ip=raw.get("ip"),
        fw=raw.get("fwver", None),
        fw_new=converters.none_or_str(raw.get("newfwver")),
        mac=raw.get("mac", None),
        ap=ap,
        parent=parent,
        state=state,
        level=level,
        config=raw.get("config"),
    )


def endpoint_update_clients(raw: str) -> dict[str, Any]:
    """Parse data from the `update_clients` endpoint"""

    parts = raw.split("originData =")
    parts = parts[1].split("networkmap_fullscan")
    data = (
        parts[0]
        .replace("fromNetworkmapd", '"fromNetworkmapd"')
        .replace("nmpClient ", '"nmpClient" ')
    )
    try:
        return json.loads(data.encode().decode("utf-8-sig"))
    except json.JSONDecodeError as ex:
        raise AsusRouterDataProcessError from ex


def pseudo_json(text: str, page: str) -> dict[str, Any]:
    """JSON parser"""

    if ENDPOINT[FIRMWARE] in page:
        return firmware(text)
    if ENDPOINT[STATE] in page:
        return {}
    if ENDPOINT[UPDATE_CLIENTS] in page:
        return endpoint_update_clients(text)
    if ENDPOINT[VPN] in page:
        return vpn_status(text)

    data = re.sub(r"\s+", "", text)

    if "curr_coreTmp" in data:
        return temperatures(data)
    # Merlin
    if ENDPOINT[SYSINFO] in page:
        return sysinfo(data)
    if ENDPOINT[ONBOARDING] in page:
        return onboarding(data)
    if "get_clientlist" in data:
        data = data.replace('"get_clientlist":', '"get_clientlist":{')
        data += "}"
    elif "get_wan_lan_status=" in data:
        data = data.replace("get_wan_lan_status=", "")
        data = data[:-1]
    elif "varcpuInfo,memInfo=newObject();cpuInfo=" in data:
        data = data.replace("varcpuInfo,memInfo=newObject();cpuInfo=", '{"cpu":{')
        data = data.replace(";memInfo=", '},"memory":{')
        data = data.replace(";", "}}")
    else:
        _LOGGER.error(
            "Unknown data. Template for this data is unknown. Endpoint=%s}", page
        )
        return {}

    try:
        return json.loads(data.encode().decode("utf-8-sig"))
    except json.JSONDecodeError as ex:
        raise AsusRouterDataProcessError from ex


def xml(text: str) -> dict[str, Any]:
    """XML parser"""

    data = xmltodict.parse(text)
    if DEVICEMAP in data:
        return devicemap(data[DEVICEMAP])

    return {}


def ovpn_client_status(raw: str) -> dict[str, Any]:
    """OpenVPN client status parser"""

    if not isinstance(raw, str):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    raw = raw.strip()
    if raw == str() or raw == "None":
        return {}

    values = {}

    for key in MAP_OVPN_CLIENT:
        if key.value in raw:
            value = re.search(f"{key.value},(.*?)(?=>)", raw)
            if value:
                values[key.get()] = key.method(value[1]) if key.method else value[1]

    return values


def ovpn_server_status(raw: str) -> dict[str, Any]:
    """OpenVPN server status parser"""

    if not isinstance(raw, str):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    raw = raw.strip()
    if raw == str() or raw == "None":
        return {}

    values = {}

    for key in MAP_OVPN_SERVER:
        if key.value in raw:
            flag = False
            # Find headers
            header = re.search(f"{AR_KEY_HEADER},{key.value},(.*?)(?=>)", raw)
            if header:
                keys_raw = header[1].split(",")
                keys = []
                for el in keys_raw:
                    keys.append(converters.to_snake_case(el))
                flag = True

            # Hide header
            raw = raw.replace(f"{AR_KEY_HEADER},{key.value},", "USED-")

            values[key.get()] = []
            # Find values
            while value := re.search(f"{key.value},(.*?)(?=>)", raw):
                if not value:
                    break
                if flag:
                    cut = value[1].split(",")
                    i = 0
                    set = {}
                    while i < len(cut) and i < len(keys):
                        set[keys[i]] = cut[i]
                        i += 1
                    values[key.get()].append(set.copy())
                else:
                    if key.method:
                        values[key.get()].append(key.method(value[1]))
                    else:
                        values[key.get()].append(value[1])
                raw = raw.replace(f"{key.value},{value[1]}", "USED-")

    return values


# var = "value";
def variables(raw: str) -> dict[str, Any]:
    """Variables parser"""

    if not isinstance(raw, str):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    values = {}
    temp = raw
    while len(temp) > 0:
        value = re.search(REGEX_VARIABLES, temp)
        if not value:
            break
        values[value[1]] = value[2]
        temp = temp.replace(f"{value[0]};", "")

    return values


def firmware(raw: str) -> dict[str, Any]:
    """Firmware parser"""

    if not isinstance(raw, str):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    values = variables(raw.replace("'", '"'))

    return values


def firmware_string(raw: str) -> Firmware | None:
    """Firmware string parser"""

    if raw == str():
        return None

    if not isinstance(raw, str):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))

    string = re.match(
        "^([39].?0.?0.?[46])?[_.]?([0-9]{3})[_.]?([0-9]+)[_.-]?([a-zA-Z0-9-_]+)?$", raw
    )
    if not string:
        _LOGGER.warning(
            "Firmware version cannot be parsed. Please report this. The original FW string is: %s",
            raw,
        )
        return Firmware()

    major = string[1]
    if major and not "." in major and len(major) == 4:
        major = major[0] + "." + major[1] + "." + major[2] + "." + major[3]

    minor = int(string[2])
    build = int(string[3])
    if not string[4]:
        build_more = 0
    elif string[4].isdigit():
        build_more = int(string[4])
    else:
        build_more = string[4]

    fw = Firmware(
        minor=minor,
        build=build,
        build_more=build_more,
    )
    if major:
        fw.major = major

    return fw


def vpn_status(raw: str) -> dict[str, Any]:
    """VPN status parser"""

    if not isinstance(raw, str):
        raise AsusRouterValueError(ERROR_VALUE_TYPE.format(raw, type(raw)))
    if raw.strip() == str():
        return {}

    vpns = {}

    values = variables(raw)

    # Clients
    for num in RANGE_OVPN_CLIENTS:
        key = f"{VPN_CLIENT}{num}_{STATUS}"
        if not key in values:
            break

        vpn_id = VPN_CLIENT + str(num)

        if values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id] = ovpn_client_status(values[key])
            vpns[vpn_id][PARAM_STATE] = True
        else:
            vpns[vpn_id] = {}
            vpns[vpn_id][PARAM_STATE] = False

        key = f"{VPN_CLIENT}{num}_{IP}"
        if key in values and values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id][PARAM_IP] = values[key]

        key = f"{VPN_CLIENT}{num}_{RIP}"
        if key in values and values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id][PARAM_RIP] = values[key]

    # Servers
    for num in RANGE_OVPN_SERVERS:
        key = f"{VPN_SERVER}{num}_{STATUS}"
        if not key in values:
            break

        vpn_id = VPN_SERVER + str(num)

        if values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id] = ovpn_server_status(values[key])
            vpns[vpn_id][PARAM_STATE] = True
        else:
            vpns[vpn_id] = {}
            vpns[vpn_id][PARAM_STATE] = False

        key = f"{VPN_SERVER}{num}_{IP}"
        if key in values and values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id][PARAM_IP] = values[key]

        key = f"{VPN_SERVER}{num}_{RIP}"
        if key in values and values[key] not in VALUES_TO_IGNORE:
            vpns[vpn_id][PARAM_RIP] = values[key]

    return vpns
